from datetime import datetime

from pydantic import BaseModel, Field

from cg.constants import FlowCellStatus
from cg.constants.devices import DeviceType
from cg.constants.metrics import DemuxMetricsColumnNames, QualityMetricsColumnNames
from cg.constants.sequencing import Sequencers


class SequencingQualityMetrics(BaseModel):
    """Model for the BCL Convert quality metrics."""

    lane: int = Field(..., alias=QualityMetricsColumnNames.LANE)
    sample_internal_id: str = Field(..., alias=QualityMetricsColumnNames.SAMPLE_INTERNAL_ID)
    mean_quality_score_q30: float = Field(
        ..., alias=QualityMetricsColumnNames.MEAN_QUALITY_SCORE_Q30
    )
    quality_score_sum: int = Field(..., alias=QualityMetricsColumnNames.QUALITY_SCORE_SUM)
    q30_bases_percent: float = Field(..., alias=QualityMetricsColumnNames.Q30_BASES_PERCENT)
    yield_: int = Field(..., alias=QualityMetricsColumnNames.YIELD)
    yield_q30: int = Field(..., alias=QualityMetricsColumnNames.YIELD_Q30)


class DemuxMetrics(BaseModel):
    """Model for the BCL Convert demultiplexing metrics."""

    lane: int = Field(..., alias=DemuxMetricsColumnNames.LANE)
    sample_internal_id: str = Field(..., alias=DemuxMetricsColumnNames.SAMPLE_INTERNAL_ID)
    read_pair_count: int = Field(..., alias=DemuxMetricsColumnNames.READ_PAIR_COUNT)


class IlluminaFlowCellDTO(BaseModel):
    """Data transfer object for Illumina flow cell."""

    internal_id: str
    type: DeviceType
    model: str

    class Config:
        arbitrary_types_allowed = True


class IlluminaSequencingRunDTO(BaseModel):
    """Data transfer object for IlluminaSequencingRun."""

    type: DeviceType
    sequencer_type: Sequencers | None
    sequencer_name: str | None
    data_availability: FlowCellStatus | None
    archived_at: datetime | None
    has_backup: bool | None
    total_reads: int | None
    total_undetermined_reads: int | None
    percent_undetermined_reads: float | None
    percent_q30: float | None
    mean_quality_score: float | None
    total_yield: int | None
    yield_q30: int | None
    cycles: int | None
    demultiplexing_software: str | None
    demultiplexing_software_version: str | None
    sequencing_started_at: datetime | None
    sequencing_completed_at: datetime | None
    demultiplexing_started_at: datetime | None
    demultiplexing_completed_at: datetime | None

    class Config:
        arbitrary_types_allowed = True


class IlluminaSampleSequencingMetricsDTO(BaseModel):
    """Data transfer object for IlluminaSampleSequencingMetrics."""

    sample_id: str
    type: DeviceType
    flow_cell_lane: int
    total_reads_in_lane: int
    base_passing_q30_percent: float
    base_mean_quality_score: float
    yield_: int
    yield_q30: float
    created_at: datetime

    class Config:
        arbitrary_types_allowed = True



from datetime import datetime

import math

from cg.constants.devices import DeviceType
from cg.services.illumina_services.illumina_metrics_service.models import (
    IlluminaSampleSequencingMetricsDTO,
)
from cg.services.illumina_services.illumina_post_processing_service.utils import (
    weighted_average,
    combine_metrics,
    combine_sample_metrics_with_undetermined,
)


def test_calculates_simple_weighted_average():
    # GIVEN Equal total counts and different percentages
    total_1, percentage_1 = 50, 0.9
    total_2, percentage_2 = 50, 0.7

    # WHEN Calculating the weighted average
    result: float = weighted_average(
        total_1=total_1, percentage_1=percentage_1, total_2=total_2, percentage_2=percentage_2
    )

    # THEN The weighted average should be 0.8
    assert math.isclose(result, 0.8, rel_tol=1e-9)


def test_handles_zero_counts():
    # GIVEN zero counts for totals
    total_1, percentage_1 = 0, 0.0
    total_2, percentage_2 = 0, 0.0

    # WHEN Calculating the weighted average
    result: float = weighted_average(
        total_1=total_1, percentage_1=percentage_1, total_2=total_2, percentage_2=percentage_2
    )

    # THEN The weighted average should be zero
    assert result == 0


def test_combine_metrics(
    mapped_metric: IlluminaSampleSequencingMetricsDTO,
    undetermined_metric: IlluminaSampleSequencingMetricsDTO,
):
    # GIVEN two metrics

    # WHEN Combining the metrics
    existing_metric: IlluminaSampleSequencingMetricsDTO = combine_metrics(
        existing_metric=mapped_metric, new_metric=undetermined_metric
    )

    # THEN The existing metric should be updated
    assert existing_metric.total_reads_in_lane == 200
    assert existing_metric.yield_ == 200


def test_combine_empty_metrics():
    # GIVEN empty lists for mapped and undetermined metrics
    mapped_metrics = []
    undetermined_metrics = []

    # WHEN combining them
    combined_metrics = combine_sample_metrics_with_undetermined(
        sample_metrics=mapped_metrics, undetermined_metrics=undetermined_metrics
    )

    # THEN the result should be an empty list
    assert combined_metrics == []


def test_combine_metrics_with_only_mapped_metrics(
    mapped_metric: IlluminaSampleSequencingMetricsDTO,
):
    # GIVEN one mapped metric and no undetermined
    mapped_metrics = [mapped_metric]
    undetermined_metrics = []

    # WHEN combining them
    combined_metrics = combine_sample_metrics_with_undetermined(
        sample_metrics=mapped_metrics, undetermined_metrics=undetermined_metrics
    )

    # THEN the result should be the mapped metrics
    assert combined_metrics == mapped_metrics


def test_combine_metrics_with_only_undetermined_metrics(
    undetermined_metric: IlluminaSampleSequencingMetricsDTO,
):
    # GIVEN an empty list of mapped metrics and list of undetermined metrics
    mapped_metrics = []
    undetermined_metrics = [undetermined_metric]

    # WHEN combining them
    combined_metrics = combine_sample_metrics_with_undetermined(
        sample_metrics=mapped_metrics, undetermined_metrics=undetermined_metrics
    )

    # THEN the result should be the undetermined metrics
    assert combined_metrics == undetermined_metrics


def test_combine_metrics_with_both_mapped_and_undetermined_metrics_different_lanes(
    mapped_metric: IlluminaSampleSequencingMetricsDTO,
    undetermined_metric: IlluminaSampleSequencingMetricsDTO,
):
    # GIVEN one mapped and one undetermined metric in different lanes for a sample
    mapped_metrics = [mapped_metric]
    undetermined_metric.flow_cell_lane = 2
    undetermined_metrics = [undetermined_metric]

    # WHEN combining them
    combined_metrics = combine_sample_metrics_with_undetermined(
        sample_metrics=mapped_metrics, undetermined_metrics=undetermined_metrics
    )

    # THEN two metrics should be returned
    assert len(combined_metrics) == 2
    assert combined_metrics[0].flow_cell_lane != combined_metrics[1].flow_cell_lane


def test_combine_metrics_with_both_mapped_and_undetermined_metrics_same_lane(
    mapped_metric: IlluminaSampleSequencingMetricsDTO,
    undetermined_metric: IlluminaSampleSequencingMetricsDTO,
):
    # GIVEN a list of mapped metrics and list of undetermined metrics in the same lane for the same sample

    # WHEN combining them
    combined_metrics = combine_sample_metrics_with_undetermined(
        sample_metrics=[mapped_metric], undetermined_metrics=[undetermined_metric]
    )

    # THEN the combined metrics should be a single metric
    assert len(combined_metrics) == 1

    # THEN the metrics should be a weighted average of the mapped and undetermined metrics
    metric: IlluminaSampleSequencingMetricsDTO = combined_metrics[0]
    assert metric.total_reads_in_lane == 200
    assert metric.yield_ == 200
    assert math.isclose(metric.yield_q30, 0.85, rel_tol=1e-9)
    assert math.isclose(metric.base_passing_q30_percent, 0.85, rel_tol=1e-9)
    assert metric.base_mean_quality_score == 25



import logging

import requests

from cg.clients.freshdesk.constants import EndPoints
from cg.clients.freshdesk.models import TicketCreate, TicketResponse

LOG = logging.getLogger(__name__)


class FreshdeskClient:
    """Client for communicating with the freshdesk REST API."""

    def __init__(self):
        self.headers = None
        self.api_key = None
        self.url = None

    def init_app(self, url: str, api_key: str):
        """Set up the client."""
        self.url = url
        self.api_key = api_key
        self.headers = {"Content-Type": "application/json"}

    @property
    def auth_header(self):
        return self.api_key, "X"

    def create_ticket(self, ticket: TicketCreate) -> TicketResponse:
        """Create a ticket."""
        LOG.debug(ticket.model_dump_json())
        response = requests.post(
            url=f"{self.url}{EndPoints.TICKETS}",
            headers=self.headers,
            auth=self.auth_header,
            json=ticket.model_dump(exclude_none=True),
        )
        return TicketResponse.model_validate(response.json())



"""Utility functions for the Illumina post-processing service."""

from pathlib import Path


from cg.constants.demultiplexing import DemultiplexingDirsAndFiles
from cg.services.illumina_services.illumina_metrics_service.models import (
    IlluminaSampleSequencingMetricsDTO,
)


def create_delivery_file_in_flow_cell_directory(flow_cell_directory: Path) -> None:
    Path(flow_cell_directory, DemultiplexingDirsAndFiles.DELIVERY).touch()


def combine_sample_metrics_with_undetermined(
    sample_metrics: list[IlluminaSampleSequencingMetricsDTO],
    undetermined_metrics: list[IlluminaSampleSequencingMetricsDTO],
) -> list[IlluminaSampleSequencingMetricsDTO]:
    """Combine metrics for sample metrics with metrics for undetermined reads.
    This occurs only if a sample is the only sample in a lane.
    """

    metrics: dict[tuple[str, int], IlluminaSampleSequencingMetricsDTO] = {
        (metric.sample_id, metric.flow_cell_lane): metric for metric in sample_metrics
    }
    for undetermined_metric in undetermined_metrics:
        key = (
            undetermined_metric.sample_id,
            undetermined_metric.flow_cell_lane,
        )
        existing_metric: IlluminaSampleSequencingMetricsDTO = metrics.get(key)

        if existing_metric:
            combined_metric: IlluminaSampleSequencingMetricsDTO = combine_metrics(
                existing_metric=existing_metric, new_metric=undetermined_metric
            )
            metrics[key] = combined_metric
        else:
            metrics[key] = undetermined_metric
    return list(metrics.values())


def combine_metrics(
    existing_metric: IlluminaSampleSequencingMetricsDTO,
    new_metric: IlluminaSampleSequencingMetricsDTO,
) -> IlluminaSampleSequencingMetricsDTO:
    """Update an existing metric with data from a new metric."""

    combined_q30_percentage: float = weighted_average(
        total_1=existing_metric.total_reads_in_lane,
        percentage_1=existing_metric.base_passing_q30_percent,
        total_2=new_metric.total_reads_in_lane,
        percentage_2=new_metric.base_passing_q30_percent,
    )
    combined_mean_quality_score: float = weighted_average(
        total_1=existing_metric.total_reads_in_lane,
        percentage_1=existing_metric.base_mean_quality_score,
        total_2=new_metric.total_reads_in_lane,
        percentage_2=new_metric.base_mean_quality_score,
    )
    combined_yield_q30_percentage: float = weighted_average(
        total_1=existing_metric.yield_,
        percentage_1=existing_metric.yield_q30,
        total_2=new_metric.yield_,
        percentage_2=new_metric.yield_q30,
    )
    combined_reads: int = existing_metric.total_reads_in_lane + new_metric.total_reads_in_lane
    combined_yield: int = existing_metric.yield_ + new_metric.yield_

    existing_metric.base_passing_q30_percent = combined_q30_percentage
    existing_metric.base_mean_quality_score = combined_mean_quality_score
    existing_metric.total_reads_in_lane = combined_reads
    existing_metric.yield_ = combined_yield
    existing_metric.yield_q30 = combined_yield_q30_percentage

    return existing_metric


def weighted_average(total_1: int, percentage_1: float, total_2: int, percentage_2: float) -> float:
    """Calculate the weighted average of two percentages."""
    if total_1 == 0 and total_2 == 0:
        return 0
    return (total_1 * percentage_1 + total_2 * percentage_2) / (total_1 + total_2)



"""Fixtures for the tests of the IlluminaPostProcessingService."""

from datetime import datetime
from pathlib import Path

import pytest
from housekeeper.store.models import Bundle

from cg.apps.housekeeper.hk import HousekeeperAPI
from cg.constants.devices import DeviceType
from cg.models.run_devices.illumina_run_directory_data import IlluminaRunDirectoryData
from cg.services.illumina_services.illumina_metrics_service.models import (
    IlluminaSampleSequencingMetricsDTO,
)
from cg.services.illumina_services.illumina_post_processing_service.illumina_post_processing_service import (
    IlluminaPostProcessingService,
)
from cg.store.models import IlluminaFlowCell
from cg.store.store import Store
from tests.store_helpers import StoreHelpers


@pytest.fixture
def illumina_post_postprocessing_service(
    store: Store,
    real_housekeeper_api: HousekeeperAPI,
    selected_novaseq_x_sample_ids: list[str],
    helpers: StoreHelpers,
) -> IlluminaPostProcessingService:
    """Return an instance of the IlluminaPostProcessingService."""
    for sample_id in selected_novaseq_x_sample_ids:
        helpers.add_sample(store, internal_id=sample_id)
    return IlluminaPostProcessingService(
        status_db=store, housekeeper_api=real_housekeeper_api, dry_run=False
    )


@pytest.fixture
def illumina_flow_cell(
    novaseq_x_demux_runs_flow_cell: IlluminaRunDirectoryData,
) -> IlluminaFlowCell:
    """Return an Illumina flow cell."""
    return IlluminaFlowCell(
        internal_id=novaseq_x_demux_runs_flow_cell.id,
        type=DeviceType.ILLUMINA,
        model=novaseq_x_demux_runs_flow_cell.run_parameters.get_flow_cell_model(),
    )


@pytest.fixture
def mapped_metric():
    return IlluminaSampleSequencingMetricsDTO(
        sample_id="sample",
        type=DeviceType.ILLUMINA,
        flow_cell_lane=1,
        total_reads_in_lane=100,
        base_passing_q30_percent=0.9,
        base_mean_quality_score=30,
        yield_=100,
        yield_q30=0.9,
        created_at=datetime.now(),
    )


@pytest.fixture
def undetermined_metric():
    return IlluminaSampleSequencingMetricsDTO(
        sample_id="sample",
        flow_cell_lane=1,
        type=DeviceType.ILLUMINA,
        total_reads_in_lane=100,
        base_passing_q30_percent=0.8,
        base_mean_quality_score=20,
        yield_=100,
        yield_q30=0.8,
        created_at=datetime.now(),
    )
